#!/usr/bin/env python3


import json
import re
import sqlite3
import subprocess
import sys
from argparse import ArgumentParser
from cmd import Cmd
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from http.client import HTTPConnection
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from ipaddress import ip_address
from os import environ
from pathlib import Path
from socket import AF_INET6
from tempfile import NamedTemporaryFile
from threading import Lock, Thread
from time import sleep
from urllib.parse import urlparse, parse_qs
from uuid import uuid4

APPDIR = Path(sys.path[0])


def sqlite_regex(pattern: str, value: str) -> bool:
    return bool(re.match(pattern, value)) if value else False


class Queries:
    INITIAL_SQL1 = r'''
    CREATE TABLE IF NOT EXISTS "database" (
        "lock"	INTEGER DEFAULT 1 CHECK("lock" = 1) UNIQUE,
        "version"	INTEGER
    );
    '''
    INITIAL_SQL2 = r'''
    CREATE TABLE IF NOT EXISTS "messages" (
        "id"	TEXT CHECK("id" REGEXP '[\da-f]{8}(-[\da-f]{4}){3}-[\da-f]{12}'),
        "sender"	TEXT NOT NULL,
        "sender_port"	INTEGER NOT NULL,
        "receiver"	TEXT NOT NULL,
        "receiver_port"	INTEGER NOT NULL,
        "time"	TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP CHECK("time" REGEXP '\d{4}-\d\d-\d\d \d\d:\d\d:\d\d'),
        "delivery_time"	TEXT CHECK("delivery_time" REGEXP '\d{4}-\d\d-\d\d \d\d:\d\d:\d\d' OR "delivery_time" IS NULL),
        "read_time"	TEXT CHECK("read_time" REGEXP '\d{4}-\d\d-\d\d \d\d:\d\d:\d\d' OR "read_time" IS NULL),
        "type"	TEXT NOT NULL DEFAULT 'text/plain',
        "status"	TEXT NOT NULL CHECK("status" IN ('waiting', 'delivered', 'read-unaware', 'read')),
        "reply_to"	TEXT,
        "value"	TEXT,
        "attachment"	TEXT CHECK("value" IS NULL OR "attachment" IS NULL),
        PRIMARY KEY("id"),
        FOREIGN KEY("reply_to") REFERENCES "messages"("id")
    );
    '''

    def __init__(self, connection):
        self.connection = connection

    def check_db_version(self) -> int:
        sql = 'SELECT version FROM database'
        if row := self.connection.execute(sql).fetchone():
            return int(row[0])

    def initialize_db(self):
        cur = self.connection.executescript(self.INITIAL_SQL1)

        if version := self.check_db_version():
            if version != 1:
                raise Exception('Incompatible database')
        else:
            cur.execute('INSERT INTO database(version) VALUES (1)')

        cur.executescript(self.INITIAL_SQL2)

    def get_manifest(self, sender_port: int, receiver: (str, int)) -> {str: dict}:
        sql = ('SELECT id, time, type, reply_to FROM messages WHERE sender="::1" AND sender_port=? AND receiver=? AND '
               'receiver_port=? AND status="waiting" ORDER BY time LIMIT 50')
        receiver_ip, receiver_port = receiver
        cur = self.connection.execute(sql, (sender_port, receiver_ip, receiver_port))
        res = {}
        while row := cur.fetchone():
            uuid, time, mime, reply_to = row
            res[uuid] = {
                'time': time,
                'type': mime,
                'reply_to': reply_to
            }
        return res

    @staticmethod
    def _convert_list(cur: sqlite3.Cursor, with_value: bool) -> {str: dict}:
        res = {}
        while row := cur.fetchone():
            uuid, time, ip, port, mime, reply_to = row[:6]
            res[uuid] = {
                'time': time,
                'party': (ip, port),
                'type': mime,
                'reply_to': reply_to
            }
            if with_value:
                res[uuid]['content'] = row[6]
        return res

    def get_pending_cli(self, sender_port: int) -> {str: dict}:
        sql = ('SELECT id, time, receiver, receiver_port, type, reply_to, value FROM messages WHERE sender="::1" AND '
               'sender_port=? AND status="waiting" ORDER BY time')
        cur = self.connection.execute(sql, (sender_port,))
        return self._convert_list(cur, True)

    def get_waiting(self, receiver_port: int) -> {str: dict}:
        sql = ('SELECT id, time, sender, sender_port, type, reply_to FROM messages WHERE receiver="::1" AND '
               'receiver_port=? AND status="waiting" ORDER BY time')
        cur = self.connection.execute(sql, (receiver_port,))
        return self._convert_list(cur, False)

    def get_pending_receivers(self, sender_port) -> [(str, int)]:
        sql = ('SELECT DISTINCT receiver, receiver_port FROM messages WHERE sender="::1" AND sender_port=? AND '
               'status="waiting"')
        cur = self.connection.execute(sql, (sender_port,))
        return cur.fetchall()

    def get_history_top(self, sender: (str, int)) -> datetime:
        sql = 'SELECT max(time) FROM messages WHERE sender=? AND sender_port=?'
        sender_ip, sender_port = sender
        if row := self.connection.execute(sql, (sender_ip, sender_port)).fetchone():
            return datetime.fromisoformat(row[0] or '1990-01-01')

    def compose_cli(self, sender_port: int, receiver: (str, int), uuid: str, content: str):
        sql = ('INSERT INTO messages(sender, sender_port, receiver, receiver_port, id, type, status, value) VALUES('
               '"::1", ?, ?, ?, ?, "text/plain", "waiting", ?)')
        receiver_ip, receiver_port = receiver
        self.connection.execute(sql, (sender_port, receiver_ip, receiver_port, uuid, content))

    def clear_pending(self, sender_port: int, receiver: (str, int) = None):
        sql1 = ('DELETE FROM messages WHERE sender="::1" AND sender_port=? AND receiver=? AND receiver_port=? AND '
                'status="waiting"')
        sql2 = 'DELETE FROM messages WHERE sender="::1" AND sender_port=? AND status="waiting"'
        if receiver:
            receiver_ip, receiver_port = receiver
            self.connection.execute(sql1, (sender_port, receiver_ip, receiver_port))
        else:
            self.connection.execute(sql2, (sender_port,))

    def add_message(self, receiver_port: int, sender: (str, int), uuid: str, params: dict):
        sql = ('INSERT INTO messages(sender, sender_port, receiver, receiver_port, id, status, time, '
               'type, reply_to) VALUES(?, ?, "::1", ?, ?, "waiting", ?, ?, ?)')
        sender_ip, sender_port = sender
        time = params['time']
        mime = params['type']
        reply_to = params.get('reply_to')
        self.connection.execute(sql, (sender_ip, sender_port, receiver_port, uuid, time, mime, reply_to))

    def find_message(self, uuid: str) -> str | None:
        sql = 'SELECT status FROM messages WHERE id=?'
        cur = self.connection.execute(sql, (uuid,))
        if row := cur.fetchone():
            return row[0]

    def get_receiver(self, uuid: str) -> (str, int):
        sql = 'SELECT receiver, receiver_port FROM messages WHERE id=?'
        cur = self.connection.execute(sql, (uuid,))
        if row := cur.fetchone():
            return row[0]


class Server(ThreadingHTTPServer):
    address_family = AF_INET6


class Handler(BaseHTTPRequestHandler):
    protocol_version = 'HTTP/1.1'

    def extract_url(self):
        url = urlparse(self.path)
        self.path = url.path
        self.params = parse_qs(url.query)

    def do_GET(self):
        self.extract_url()
        ip = self.client_address[0]

        if self.path.startswith('/api-v1/'):
            match self.path[len('/api-v1/'):]:
                case 'request-sync':
                    # this code is handled by the receiver
                    port = self.params.get('from', [''])[0]  # sender's port
                    set_proof = self.headers.get('Set-Proof')

                    if port.isnumeric():
                        port = int(port)
                    else:
                        return self.send_error(400,
                                               explain="Invalid param 'from'")

                    if bool(pref.get('block-foreign')) and not Person(host=(ip, port)).name:
                        return self.send_error(403,
                                               explain='Communication with the host is disabled')

                    if not set_proof:
                        return self.send_error(400,
                                               explain="Missing Set-Proof header")

                    if (ip, port) in hosts.keys():
                        hosts[(ip, port)].update(set_proof, False)
                    else:
                        hosts[(ip, port)] = Host((ip, port), set_proof, False)

                    self.send_response(200)
                    self.send_header('Content-Length', 0)
                    self.end_headers()

                    executor.submit(sync_manifest, ip, port)

                case 'sync':
                    # this code is handled by the sender
                    port = self.headers.get('To')  # receiver's port
                    proof = self.headers.get('Proof')
                    uuid = self.params.get('uuid', [''])[0]

                    if None in (port, proof):
                        return self.send_error(400,
                                               explain="Some of the required headers are missing")

                    if port.isnumeric():
                        port = int(port)
                    else:
                        return self.send_error(400,
                                               explain="Invalid To header")

                    if (ip, port) in hosts.keys() and hosts[(ip, port)].verify(proof):
                        if uuid:
                            ...
                        else:
                            waiting = queries.get_manifest(args.port, (ip, port))
                            payload = json.dumps(waiting, indent=4).encode()
                            self.send_response(200)
                            self.send_header('Content-Type', 'application/json')
                            self.send_header('Content-Length', len(payload))
                            self.end_headers()
                            self.wfile.write(payload)
                    else:
                        return self.send_error(403,
                                               explain='Authentication failed')

                case _:
                    self.send_error(404)
        else:
            self.send_error(404)


class Host:
    def __init__(self, host: (str, int), proof: str, trusted: bool):
        self.host = host
        self.proof = proof
        self.is_trusted = trusted
        self.old_proof = None
        self.lock = Lock()

    def __repr__(self):
        uid = Person(host=self.host).display
        proofs = self.proof
        if not self.is_trusted:
            proofs += '*'
        if self.old_proof:
            proofs = f'[{self.old_proof}] {proofs}'
        return uid + ' ' + proofs

    def verify(self, proof: str) -> bool:
        if proof is None:
            return False

        if proof == self.proof:
            self.lock.acquire()
            self.is_trusted = True
            self.old_proof = None
            self.lock.release()
            return True

        elif proof == self.old_proof:
            self.lock.acquire()
            self.proof = self.old_proof
            self.is_trusted = True
            self.old_proof = None
            self.lock.release()
            return True

        else:
            return False

    def update(self, proof: str, trusted: bool):
        self.lock.acquire()
        if trusted:
            self.old_proof = None

        elif self.is_trusted and not trusted:
            self.old_proof = self.proof

        self.is_trusted = trusted
        self.proof = proof
        self.lock.release()


class Person:
    def __init__(self, /, host: (str, int) = None, addr: str = None, name: str = None, display: str = None):
        if display:  # display -> name | display -> addr
            if display in pref.get('contacts', {}).values():
                name = display
            else:
                addr = display

        if name:  # name -> addr
            if 'contacts' in pref:
                addr = list(pref['contacts'])[
                    list(pref['contacts'].values()).index(name)
                ]

        if addr:  # addr -> host
            host = addr.split('|', 1)

        if host:  # host -> ip, port
            ip = str(ip_address(host[0]))
            port = int(host[1])
            host = ip, port

            if not addr:  # host -> addr
                addr = '%s|%i' % (ip, port)

            if not name:  # addr -> name
                name = pref.get('contacts', {}).get(addr)

        self.host = host
        self.addr = addr
        self.name = name
        self.display = name or addr


# jobs -->

def sync_message(uuid: str, host: (str, int) = None, params: dict = None):
    print('sync message', uuid)

    if not host or not params:
        host = queries.get_receiver(uuid)
        if not host:
            return

    ip, port = host
    host = hosts[(ip, port)]

    try:
        req = HTTPConnection(ip, port, timeout=10)
        req.request('GET', '/api-v1/sync?uuid=%s' % uuid, headers={
            'To': args.port,
            'Proof': host.proof
        })
        res = req.getresponse()
    except Exception as e:
        print('sync message', type(e).__name__, e)
        return

    if res.status == 200:
        host.verify(host.proof)
    else:
        print('sync message', ip, port, 'out of sync')
        return

    size = int(res.getheader('Content-Length'))


def sync_manifest(ip: str, port: int):
    # fetching manifest
    print('sync manifest', ip, port)

    host = hosts[(ip, port)]

    try:
        req = HTTPConnection(ip, port, timeout=10)
        req.request('GET', '/api-v1/sync', headers={
            'To': args.port,
            'Proof': host.proof
        })
        res = req.getresponse()

    except Exception as e:
        print('sync manifest', type(e).__name__, e)
        return

    if res.status == 200:
        host.verify(host.proof)
    else:
        print('sync manifest', ip, port, 'out of sync')
        return

    size = int(res.getheader('Content-Length'))
    data = res.read(min(size, 300 * 50))
    manifest = json.loads(data)

    # prevent history injection
    history_top = queries.get_history_top((ip, port))
    print('sync manifest', ip, port, history_top)

    for uuid, params in manifest.items():
        print('sync manifest', ip, port, uuid, params)
        status = queries.find_message(uuid)

        if datetime.fromisoformat(params['time']) < history_top:
            print('sync manifest', uuid, 'history injection')
            continue

        if status not in (None, 'waiting'):
            print('sync manifest', uuid, 'conflict')
            continue

        if params.get('reply_to') and queries.find_message(params['reply_to']):
            print('sync manifest', uuid, 'reply-to not found')
            continue

        if status is None:
            queries.add_message(args.port, (ip, port), uuid, params)

        executor.submit(sync_message, uuid, (ip, port), params)
        print('sync manifest', 'Message awaits fetching')
    con.commit()


def request_sync(ip: str, port: int):
    print('request sync', ip, port)
    proof = str(uuid4())
    if (ip, port) in hosts:
        hosts[(ip, port)].update(proof, True)
    else:
        hosts[(ip, port)] = Host((ip, port), proof, True)

    try:
        req = HTTPConnection(ip, port, timeout=10)
        req.request('GET', '/api-v1/request-sync?from=%i' % args.port, headers={
            'Set-Proof': proof
        })
        res = req.getresponse()

    except Exception as e:
        print('request sync', type(e).__name__, e)
        return

    if res.status == 200:
        print('request sync success', ip, port)
    else:
        print('request sync fail', ip, port, res.status)


# <-- jobs


class ServerConsole(Cmd):
    prompt = ''

    def do_send(self, dest=''):
        """send a message
send
send Alice
send ::1|2024
"""

        with NamedTemporaryFile(mode='w+') as tf:
            tf.write(f'''To: {dest}
# Message content goes below this line #
''')
            tf.flush()

            while True:
                subprocess.call((environ.get('EDITOR', 'vim'), tf.name))
                tf.seek(0)
                lines = []
                data_flag = valid_flag = False

                while line := tf.readline().strip('\n'):
                    if data_flag:
                        lines.append(line)
                        continue

                    if line.startswith('#'):
                        data_flag = True
                    elif line.startswith('To: '):
                        dest = line.split(':', 1)[1].strip()

                        try:
                            ip, port = Person(display=dest).host
                        except (ValueError, TypeError):
                            print('Wrong destination', file=sys.stderr)
                            break
                        else:
                            valid_flag = True

                if valid_flag:
                    break
                elif input('Cancel? [y/N] ').lower() == 'y':
                    return

        if data := '\n'.join(lines).strip():
            print('Putting your message into the queue', file=sys.stderr)
            queries.compose_cli(args.port, (ip, port), str(uuid4()), data)
            con.commit()

            executor.submit(request_sync, ip, port)

    def do_pending(self, arg):
        """Show outgoing message queue"""

        for uuid, message in queries.get_pending_cli(args.port).items():
            print(f'-----{uuid}-----')
            for key, value in message.items():
                if key == 'content':
                    print()
                    print(value)
                    print()
                elif key == 'party':
                    print('receiver:', Person(host=value).display)
                elif value is not None:
                    print(key + ':', value)

    def do_waiting(self, arg):
        """Show incoming message queue"""

        for uuid, message in queries.get_waiting(args.port).items():
            print(f'-----{uuid}-----')
            for key, value in message.items():
                if key == 'content':
                    print()
                    print(value)
                    print()
                elif key == 'party':
                    print('sender:', Person(host=value).display)

    def do_clear(self, dest):
        """Clear message queue
clear
clear Alice
clear ::1|2024"""

        if dest:
            try:
                dest = Person(name=dest).host
            except ValueError:
                print('Wrong destination', file=sys.stderr)
                return
            queries.clear_pending(args.port, dest)
        else:
            queries.clear_pending(args.port)

        con.commit()

    def do_hosts(self, arg):
        """Show hosts and its proofs"""

        for host in hosts.values():
            print(host)

    def do_sync(self, arg):
        """Sync message queue"""

        for receiver in queries.get_pending_receivers(args.port):
            executor.submit(request_sync, *receiver)


hosts = {}

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('--conf', default=APPDIR / 'pref.json', type=Path)
    parser.add_argument('--db', default=APPDIR / 'chats.db', type=Path)
    parser.add_argument('--port', default=2024, type=int)
    args = parser.parse_args()

    with open(args.conf) as fd:
        pref = json.load(fd)

    if pref['version'] != 1:
        raise Exception('Incompatible config file')

    con = sqlite3.connect(args.db, check_same_thread=False)
    con.create_function('regexp', 2, sqlite_regex)
    queries = Queries(con)
    queries.initialize_db()

    with Server(('::', args.port), Handler) as server:
        print(f'http://[::]:{args.port}', file=sys.stderr)
        executor = ThreadPoolExecutor(max_workers=10)
        Thread(target=ServerConsole().cmdloop, daemon=True).start()
        server.serve_forever()
